from ..utils import stats
import numpy as np
from numpy import ma
np.random.seed(10031312)


def test_median_axis_none_mask_none():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        expected = np.median(a.astype(np.float32))
        actual = stats.median(a)
        assert np.float32(expected) == actual


def test_median_2d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(a.astype(np.float32))
        actual = stats.median(a)
        assert np.float32(expected) == actual


def test_median_3d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(a.astype(np.float32))
        actual = stats.median(a)
        assert np.float32(expected) == actual


def test_median_2d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(a.astype(np.float32), axis=0)
        actual = stats.median(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_2d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(a.astype(np.float32), axis=1)
        actual = stats.median(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(a.astype(np.float32), axis=0)
        actual = stats.median(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(a.astype(np.float32), axis=1)
        actual = stats.median(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_2_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(a.astype(np.float32), axis=2)
        actual = stats.median(a, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_axis_none_mask():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        value_to_mask = np.random.uniform(0, 1.0)
        mask = np.random.uniform(0, 1, size) < value_to_mask
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32))
        actual = stats.median(a, mask=mask)
        assert np.float32(expected) == actual


def test_median_2d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0, 1, size=(size1, size2)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32))
        actual = stats.median(a, mask=mask)
        assert np.float32(expected) == actual


def test_median_3d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0., 1.0, size=(size1, size2, size3)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32))
        actual = stats.median(a, mask=mask)
        assert np.float32(expected) == actual


def test_median_2d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0., 1.0, size=(size1, size2)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32), axis=0)
        actual = stats.median(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_2d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0., 1.0, size=(size1, size2)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32), axis=1)
        actual = stats.median(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0, 1, size=(size1, size2, size3)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32), axis=0)
        actual = stats.median(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0, 1, size=(size1, size2, size3)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32), axis=1)
        actual = stats.median(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_median_3d_axis_2_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        value_to_mask = np.random.uniform(0, 1)
        mask = np.random.uniform(0, 1, size=(size1, size2, size3)) < value_to_mask
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = ma.median(ma.array(a, mask=mask, dtype=np.float32), axis=2)
        actual = stats.median(a, mask=mask, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-6)


def test_absolute_deviation_axis_none_mask_none():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32)))
        actual = stats.absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_2d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32)))
        actual = stats.absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_3d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32)))
        actual = stats.absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_2d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=0))
        actual = stats.absolute_deviation(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_absolute_deviation_2d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.abs(a.astype(np.float32).T - np.median(a.astype(np.float32), axis=1)).T
        actual = stats.absolute_deviation(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_absolute_deviation_3d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=0))
        actual = stats.absolute_deviation(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_absolute_deviation_3d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=1, keepdims=True))
        actual = stats.absolute_deviation(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_absolute_deviation_3d_axis_2_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=2, keepdims=True))
        actual = stats.absolute_deviation(a, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_absolute_deviation_axis_none_mask():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a_masked - ma.median(a_masked))
        actual = stats.absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_2d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - ma.median(a_masked))
        actual = stats.absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_3d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - ma.median(a_masked))
        actual = stats.absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-8)


def test_absolute_deviation_2d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - ma.median(a_masked, axis=0))
        actual = stats.absolute_deviation(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_2d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a.T - ma.median(a_masked, axis=1)).T
        actual = stats.absolute_deviation(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_absolute_deviation_3d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - ma.median(a_masked, axis=0))
        actual = stats.absolute_deviation(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


def test_absolute_deviation_3d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - np.expand_dims(ma.median(a_masked, axis=1), axis=1))
        actual = stats.absolute_deviation(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


def test_absolute_deviation_3d_axis_2_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = np.abs(a - np.expand_dims(ma.median(a_masked, axis=2), axis=2))
        actual = stats.absolute_deviation(a, mask=mask, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


def test_mad_axis_none_mask_none():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32))))
        actual = stats.median_absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_2d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32))))
        actual = stats.median_absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_3d_axis_none_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32))))
        actual = stats.median_absolute_deviation(a)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_2d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=0)), axis=0)
        actual = stats.median_absolute_deviation(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_mad_2d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        expected = np.median(np.abs(a.astype(np.float32).T - np.median(a.astype(np.float32), axis=1)).T, axis=1)
        actual = stats.median_absolute_deviation(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_mad_3d_axis_0_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=0)), axis=0)
        actual = stats.median_absolute_deviation(a, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_mad_3d_axis_1_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=1, keepdims=True)), axis=1)
        actual = stats.median_absolute_deviation(a, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_mad_3d_axis_2_mask_none():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        expected = np.median(np.abs(a.astype(np.float32) - np.median(a.astype(np.float32), axis=2, keepdims=True)), axis=2)
        actual = stats.median_absolute_deviation(a, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-4)


def test_mad_axis_none_mask():
    for i in range(1000):
        size = np.random.randint(1, 10000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size)
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a_masked - ma.median(a_masked)), dtype=np.float32, mask=mask))
        actual = stats.median_absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_2d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - ma.median(a_masked)), dtype=np.float32, mask=mask))
        actual = stats.median_absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, expected, atol=1e-4)


def test_mad_3d_axis_none_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - ma.median(a_masked)), dtype=np.float32, mask=mask))
        actual = stats.median_absolute_deviation(a, mask=mask)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-8)


def test_mad_2d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - ma.median(a_masked, axis=0)), dtype=np.float32, mask=mask), axis=0)
        actual = stats.median_absolute_deviation(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_2d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 1000)
        size2 = np.random.randint(1, 1000)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a.T - ma.median(a_masked, axis=1)).T, dtype=np.float32, mask=mask), axis=1)
        actual = stats.median_absolute_deviation(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, np.float32(expected), atol=1e-4)


def test_mad_3d_axis_0_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - ma.median(a_masked, axis=0)), dtype=np.float32, mask=mask), axis=0)
        actual = stats.median_absolute_deviation(a, mask=mask, axis=0)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


def test_mad_3d_axis_1_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - np.expand_dims(ma.median(a_masked, axis=1), axis=1)), dtype=np.float32, mask=mask), axis=1)
        actual = stats.median_absolute_deviation(a, mask=mask, axis=1)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


def test_mad_3d_axis_2_mask():
    for i in range(10):
        size1 = np.random.randint(1, 100)
        size2 = np.random.randint(1, 100)
        size3 = np.random.randint(1, 100)
        mean = np.random.uniform(-1000, 1000)
        sigma = np.random.uniform(0, 1000)
        a = np.random.normal(mean, sigma, size=(size1, size2, size3))
        value_to_mask = np.random.uniform(0, 0.8)
        mask = np.random.uniform(0, 1.0, size=(size1, size2, size3)) < value_to_mask
        a_masked = ma.array(a, mask=mask, dtype=np.float32)
        expected = ma.median(ma.array(np.abs(a - np.expand_dims(ma.median(a_masked, axis=2), axis=2)), dtype=np.float32, mask=mask), axis=2)
        actual = stats.median_absolute_deviation(a, mask=mask, axis=2)
        np.testing.assert_allclose(actual, expected.astype(np.float32), atol=1e-9)


# def test_rstd_axis_none_mask_none():
#     for i in range(1000):
#         size = np.random.randint(1, 10000)
#         mean = np.random.uniform(-1000, 1000)
#         sigma = np.random.uniform(0, 1000)
#         a = np.random.normal(mean, sigma, size)
#         expected = np.std(a)
#         actual = stats.robust_standard_deviation(a)
#         np.testing.assert_allclose(actual, np.float32(expected), atol= sigma / (size ** 0.5))



from ..bias import BiasSubtractor
from .utils import FakeImage, throws_inhomogeneous_set_exception
from ..stages import MasterCalibrationDoesNotExist
import pytest
import mock

import numpy as np


class FakeBiasImage(FakeImage):
    def __init__(self, *args, bias_level=0.0, **kwargs):
        super(FakeBiasImage, self).__init__(*args, image_multiplier=bias_level, **kwargs)
        self.header = {'BIASLVL': bias_level}


def test_no_input_images():
    subtractor = BiasSubtractor(None)
    images = subtractor.do_stage([])
    assert len(images) == 0


def test_group_by_keywords():
    subtractor = BiasSubtractor(None)
    assert subtractor.group_by_keywords == ['ccdsum']


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_header_has_biaslevel(mock_cal, mock_image):
    mock_image.return_value = FakeBiasImage()
    subtractor = BiasSubtractor(None)
    images = subtractor.do_stage([FakeImage() for x in range(6)])
    for image in images:
        assert image.header['BIASLVL'][0] == 0


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_header_biaslevel_is_1(mock_cal, mock_image):
    mock_image.return_value = FakeBiasImage(bias_level=1)
    subtractor = BiasSubtractor(None)
    images = subtractor.do_stage([FakeImage() for x in range(6)])
    for image in images:
        assert image.header['BIASLVL'][0] == 1


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_header_biaslevel_is_2(mock_cal, mock_image):
    mock_image.return_value = FakeBiasImage(bias_level=2.0)
    subtractor = BiasSubtractor(None)
    images = subtractor.do_stage([FakeImage() for x in range(6)])
    for image in images:
        assert image.header['BIASLVL'][0] == 2


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_raises_an_exection_if_ccdsums_are_different(mock_cal, mock_images):
    throws_inhomogeneous_set_exception(BiasSubtractor, None, 'ccdsum', '1 1')


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_raises_an_exection_if_epochs_are_different(mock_cal, mock_images):
    throws_inhomogeneous_set_exception(BiasSubtractor, None, 'epoch', '20160102')


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_raises_an_exection_if_nx_are_different(mock_cal, mock_images):
    mock_cal.return_value = 'test.fits'
    throws_inhomogeneous_set_exception(BiasSubtractor, None, 'nx', 105)


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_raises_an_exection_if_ny_are_different(mock_cal, mock_images):
    mock_cal.return_value = 'test.fits'
    throws_inhomogeneous_set_exception(BiasSubtractor, None, 'ny', 107)


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_raises_exception_if_no_master_calibration(mock_cal, mock_images):
    mock_cal.return_value = None
    mock_images.return_value = FakeBiasImage()
    subtractor = BiasSubtractor(None)

    with pytest.raises(MasterCalibrationDoesNotExist):
        images = subtractor.do_stage([FakeImage() for x in range(6)])


@mock.patch('banzai.stages.Image')
@mock.patch('banzai.stages.ApplyCalibration.get_calibration_filename')
def test_bias_subtraction_is_reasonable(mock_cal, mock_image):
    mock_cal.return_value = 'test.fits'
    input_bias = 1000.0
    input_readnoise = 9.0
    input_level = 2000.0
    nx = 101
    ny = 103

    fake_master_bias = FakeBiasImage(bias_level=input_bias)
    fake_master_bias.data = np.random.normal(0.0, input_readnoise, size=(ny, nx))
    mock_image.return_value = fake_master_bias

    subtractor = BiasSubtractor(None)
    images = [FakeImage(image_multiplier=input_level) for x in range(6)]

    images = subtractor.do_stage(images)

    for image in images:
        assert np.abs(image.header['BIASLVL'][0] - input_bias) < 1.0
        assert np.abs(np.mean(image.data) - input_level + input_bias) < 1.0



"""
Old implementation of ``./setup.py test`` command.  This has been moved to
astropy.tests as of Astropy v1.1.0, but a copy of the implementation is kept
here for backwards compatibility.
"""

from __future__ import absolute_import, unicode_literals

import inspect
import os
import shutil
import subprocess
import sys
import tempfile

from setuptools import Command

from ..compat import _fix_user_options


PY3 = sys.version_info[0] == 3


class AstropyTest(Command, object):
    description = 'Run the tests for this package'

    user_options = [
        ('package=', 'P',
         "The name of a specific package to test, e.g. 'io.fits' or 'utils'.  "
         "If nothing is specified, all default tests are run."),
        ('test-path=', 't',
         'Specify a test location by path.  If a relative path to a  .py file, '
         'it is relative to the built package, so e.g., a  leading "astropy/" '
         'is necessary.  If a relative  path to a .rst file, it is relative to '
         'the directory *below* the --docs-path directory, so a leading '
         '"docs/" is usually necessary.  May also be an absolute path.'),
        ('verbose-results', 'V',
         'Turn on verbose output from pytest.'),
        ('plugins=', 'p',
         'Plugins to enable when running pytest.'),
        ('pastebin=', 'b',
         "Enable pytest pastebin output. Either 'all' or 'failed'."),
        ('args=', 'a',
         'Additional arguments to be passed to pytest.'),
        ('remote-data', 'R', 'Run tests that download remote data.'),
        ('pep8', '8',
         'Enable PEP8 checking and disable regular tests. '
         'Requires the pytest-pep8 plugin.'),
        ('pdb', 'd',
         'Start the interactive Python debugger on errors.'),
        ('coverage', 'c',
         'Create a coverage report. Requires the coverage package.'),
        ('open-files', 'o', 'Fail if any tests leave files open.  Requires the '
         'psutil package.'),
        ('parallel=', 'j',
         'Run the tests in parallel on the specified number of '
         'CPUs.  If negative, all the cores on the machine will be '
         'used.  Requires the pytest-xdist plugin.'),
        ('docs-path=', None,
         'The path to the documentation .rst files.  If not provided, and '
         'the current directory contains a directory called "docs", that '
         'will be used.'),
        ('skip-docs', None,
         "Don't test the documentation .rst files."),
        ('repeat=', None,
         'How many times to repeat each test (can be used to check for '
         'sporadic failures).'),
        ('temp-root=', None,
         'The root directory in which to create the temporary testing files. '
         'If unspecified the system default is used (e.g. /tmp) as explained '
         'in the documentation for tempfile.mkstemp.')
    ]

    user_options = _fix_user_options(user_options)

    package_name = ''

    def initialize_options(self):
        self.package = None
        self.test_path = None
        self.verbose_results = False
        self.plugins = None
        self.pastebin = None
        self.args = None
        self.remote_data = False
        self.pep8 = False
        self.pdb = False
        self.coverage = False
        self.open_files = False
        self.parallel = 0
        self.docs_path = None
        self.skip_docs = False
        self.repeat = None
        self.temp_root = None

    def finalize_options(self):
        # Normally we would validate the options here, but that's handled in
        # run_tests
        pass

    # Most of the test runner arguments have the same name as attributes on
    # this command class, with one exception (for now)
    _test_runner_arg_attr_map = {
        'verbose': 'verbose_results'
    }

    def generate_testing_command(self):
        """
        Build a Python script to run the tests.
        """

        cmd_pre = ''  # Commands to run before the test function
        cmd_post = ''  # Commands to run after the test function

        if self.coverage:
           pre, post = self._generate_coverage_commands()
           cmd_pre += pre
           cmd_post += post

        def get_attr(arg):
            attr = self._test_runner_arg_attr_map.get(arg, arg)
            return getattr(self, attr)

        test_args = filter(lambda arg: hasattr(self, arg),
                           self._get_test_runner_args())

        test_args = ', '.join('{0}={1!r}'.format(arg, get_attr(arg))
                              for arg in test_args)

        if PY3:
            set_flag = "import builtins; builtins._ASTROPY_TEST_ = True"
        else:
            set_flag = "import __builtin__; __builtin__._ASTROPY_TEST_ = True"

        cmd = ('{cmd_pre}{0}; import {1.package_name}, sys; result = '
               '{1.package_name}.test({test_args}); {cmd_post}'
               'sys.exit(result)')

        return cmd.format(set_flag, self, cmd_pre=cmd_pre,
                          cmd_post=cmd_post, test_args=test_args)

    def _validate_required_deps(self):
        """
        This method checks that any required modules are installed before
        running the tests.
        """
        try:
            import astropy
        except ImportError:
            raise ImportError(
                "The 'test' command requires the astropy package to be "
                "installed and importable.")

    def run(self):
        """
        Run the tests!
        """

        # Ensure there is a doc path
        if self.docs_path is None:
            if os.path.exists('docs'):
                self.docs_path = os.path.abspath('docs')

        # Build a testing install of the package
        self._build_temp_install()

        # Ensure all required packages are installed
        self._validate_required_deps()

        # Run everything in a try: finally: so that the tmp dir gets deleted.
        try:
            # Construct this modules testing command
            cmd = self.generate_testing_command()
            # Run the tests in a subprocess--this is necessary since
            # new extension modules may have appeared, and this is the
            # easiest way to set up a new environment

            # On Python 3.x prior to 3.3, the creation of .pyc files
            # is not atomic.  py.test jumps through some hoops to make
            # this work by parsing import statements and carefully
            # importing files atomically.  However, it can't detect
            # when __import__ is used, so its carefulness still fails.
            # The solution here (admittedly a bit of a hack), is to
            # turn off the generation of .pyc files altogether by
            # passing the `-B` switch to `python`.  This does mean
            # that each core will have to compile .py file to bytecode
            # itself, rather than getting lucky and borrowing the work
            # already done by another core.  Compilation is an
            # insignificant fraction of total testing time, though, so
            # it's probably not worth worrying about.
            retcode = subprocess.call([sys.executable, '-B', '-c', cmd],
                                      cwd=self.testing_path, close_fds=False)

        finally:
            # Remove temporary directory
            shutil.rmtree(self.tmp_dir)

        raise SystemExit(retcode)

    def _build_temp_install(self):
        """
        Build the package and copy the build to a temporary directory for
        the purposes of testing this avoids creating pyc and __pycache__
        directories inside the build directory
        """
        self.reinitialize_command('build', inplace=True)
        self.run_command('build')
        build_cmd = self.get_finalized_command('build')
        new_path = os.path.abspath(build_cmd.build_lib)

        # On OSX the default path for temp files is under /var, but in most
        # cases on OSX /var is actually a symlink to /private/var; ensure we
        # dereference that link, because py.test is very sensitive to relative
        # paths...
        tmp_dir = tempfile.mkdtemp(prefix=self.package_name + '-test-',
                                   dir=self.temp_root)
        self.tmp_dir = os.path.realpath(tmp_dir)
        self.testing_path = os.path.join(self.tmp_dir, os.path.basename(new_path))
        shutil.copytree(new_path, self.testing_path)

        new_docs_path = os.path.join(self.tmp_dir,
                                     os.path.basename(self.docs_path))
        shutil.copytree(self.docs_path, new_docs_path)
        self.docs_path = new_docs_path

        shutil.copy('setup.cfg', self.tmp_dir)

    def _generate_coverage_commands(self):
        """
        This method creates the post and pre commands if coverage is to be
        generated
        """
        if self.parallel != 0:
            raise ValueError(
                "--coverage can not be used with --parallel")

        try:
            import coverage
        except ImportError:
            raise ImportError(
                "--coverage requires that the coverage package is "
                "installed.")

        # Don't use get_pkg_data_filename here, because it
        # requires importing astropy.config and thus screwing
        # up coverage results for those packages.
        coveragerc = os.path.join(
            self.testing_path, self.package_name, 'tests', 'coveragerc')

        # We create a coveragerc that is specific to the version
        # of Python we're running, so that we can mark branches
        # as being specifically for Python 2 or Python 3
        with open(coveragerc, 'r') as fd:
            coveragerc_content = fd.read()
        if PY3:
            ignore_python_version = '2'
        else:
            ignore_python_version = '3'
        coveragerc_content = coveragerc_content.replace(
            "{ignore_python_version}", ignore_python_version).replace(
                "{packagename}", self.package_name)
        tmp_coveragerc = os.path.join(self.tmp_dir, 'coveragerc')
        with open(tmp_coveragerc, 'wb') as tmp:
            tmp.write(coveragerc_content.encode('utf-8'))

        cmd_pre = (
            'import coverage; '
            'cov = coverage.coverage(data_file="{0}", config_file="{1}"); '
            'cov.start();'.format(
                os.path.abspath(".coverage"), tmp_coveragerc))
        cmd_post = (
            'cov.stop(); '
            'from astropy.tests.helper import _save_coverage; '
            '_save_coverage(cov, result, "{0}", "{1}");'.format(
                os.path.abspath('.'), self.testing_path))

        return cmd_pre, cmd_post

    def _get_test_runner_args(self):
        """
        A hack to determine what arguments are supported by the package's
        test() function.  In the future there should be a more straightforward
        API to determine this (really it should be determined by the
        ``TestRunner`` class for whatever version of Astropy is in use).
        """

        if PY3:
            import builtins
            builtins._ASTROPY_TEST_ = True
        else:
            import __builtin__
            __builtin__._ASTROPY_TEST_ = True

        try:
            pkg = __import__(self.package_name)
            if not hasattr(pkg, 'test'):
                raise ImportError(
                    'package {0} does not have a {0}.test() function as '
                    'required by the Astropy test runner'.format(package_name))

            argspec = inspect.getargspec(pkg.test)
            return argspec.args
        finally:
            if PY3:
                del builtins._ASTROPY_TEST_
            else:
                del __builtin__._ASTROPY_TEST_



from __future__ import absolute_import, print_function, division

import numpy as np
import os.path

from .utils import stats, fits_utils
from banzai import logs
from banzai.images import Image
from .stages import CalibrationMaker, ApplyCalibration

__author__ = 'cmccully'


class DarkMaker(CalibrationMaker):
    def __init__(self, pipeline_context):
        super(DarkMaker, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return ['ccdsum']

    @property
    def calibration_type(self):
        return 'DARK'

    @property
    def min_images(self):
        return 5

    def make_master_calibration_frame(self, images, image_config, logging_tags):
        dark_data = np.zeros((images[0].ny, images[0].nx, len(images)), dtype=np.float32)
        dark_mask = np.zeros((images[0].ny, images[0].nx, len(images)), dtype=np.uint8)

        master_dark_filename = self.get_calibration_filename(images[0])

        logs.add_tag(logging_tags, 'master_dark', os.path.basename(master_dark_filename))
        for i, image in enumerate(images):
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            self.logger.debug('Combining dark', extra=logging_tags)

            dark_data[:, :, i] = image.data[:, :]
            dark_data[:, :, i] /= image.exptime
            dark_mask[:, :, i] = image.bpm[:, :]

        master_dark = stats.sigma_clipped_mean(dark_data, 3.0, axis=2, mask=dark_mask, inplace=True)

        # Memory cleanup
        del dark_data
        del dark_mask

        master_bpm = np.array(master_dark == 0.0, dtype=np.uint8)
        master_dark[master_bpm] = 0.0

        # Save the master dark image with all of the combined images in the header
        master_dark_header = fits_utils.create_master_calibration_header(images)
        master_dark_image = Image(data=master_dark, header=master_dark_header)
        master_dark_image.filename = master_dark_filename
        master_dark_image.bpm = master_bpm

        logs.pop_tag(logging_tags, 'master_dark')
        logs.add_tag(logging_tags, 'filename', os.path.basename(master_dark_image.filename))
        self.logger.info('Created master dark', extra=logging_tags)
        return [master_dark_image]


class DarkSubtractor(ApplyCalibration):
    def __init__(self, pipeline_context):
        super(DarkSubtractor, self).__init__(pipeline_context)

    @property
    def calibration_type(self):
        return 'dark'

    @property
    def group_by_keywords(self):
        return ['ccdsum']

    def apply_master_calibration(self, images, master_calibration_image, logging_tags):
        master_dark_data = master_calibration_image.data
        master_dark_filename = os.path.basename(master_calibration_image.filename)
        logs.add_tag(logging_tags, 'master_dark',
                     os.path.basename(master_calibration_image.filename))

        for image in images:
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            self.logger.info('Subtracting dark', extra=logging_tags)
            image.data -= master_dark_data * image.exptime
            image.bpm |= master_calibration_image.bpm
            image.header['L1IDDARK'] = (master_dark_filename, 'ID of dark frame used')
            image.header['L1STATDA'] = (1, 'Status flag for dark frame correction')
        return images



from ..bias import OverscanSubtractor
from .utils import FakeImage

import numpy as np


class FakeOverscanImage(FakeImage):
    def __init__(self, *args, **kwargs):
        super(FakeOverscanImage, self).__init__(*args, **kwargs)
        self.header = {'BIASSEC': 'UNKNOWN'}


def test_no_input_images():
    subtractor = OverscanSubtractor(None)
    images = subtractor.do_stage([])
    assert len(images) == 0


def test_group_by_keywords():
    subtractor = OverscanSubtractor(None)
    assert subtractor.group_by_keywords is None


def test_header_has_overscan_when_biassec_unknown():
    subtractor = OverscanSubtractor(None)
    images = subtractor.do_stage([FakeOverscanImage() for x in range(6)])
    for image in images:
        assert image.header['OVERSCAN'][0] == 0


def test_header_overscan_is_1():
    subtractor = OverscanSubtractor(None)
    nx = 101
    ny = 103
    noverscan = 10
    images = [FakeOverscanImage(nx=nx, ny=ny) for x in range(6)]
    for image in images:
        image.header['BIASSEC'] = '[{nover}:{nx},1:{ny}]'.format(nover=noverscan, nx=nx, ny=ny)
    images = subtractor.do_stage(images)
    for image in images:
        assert image.header['OVERSCAN'][0] == 1


def test_header_overscan_is_2():
    subtractor = OverscanSubtractor(None)
    nx = 101
    ny = 103
    noverscan = 10
    images =[FakeOverscanImage(nx=nx, ny=ny, image_multiplier=2) for x in range(6)]
    for image in images:
        image.header['BIASSEC'] = '[{nover}:{nx},1:{ny}]'.format(nover=nx-noverscan, nx=nx, ny=ny)
    images = subtractor.do_stage(images)
    for image in images:
        assert image.header['OVERSCAN'][0] == 2


def test_overscan_estimation_is_reasonable():
    subtractor = OverscanSubtractor(None)
    nx = 101
    ny = 103
    noverscan = 10
    expected_read_noise = 10.0
    expected_overscan = 40.0
    input_level = 100

    images =[FakeOverscanImage(nx=nx, ny=ny, image_multiplier=input_level) for x in range(6)]
    for image in images:
        image.header['BIASSEC'] = '[{nover}:{nx},1:{ny}]'.format(nover=nx - noverscan + 1,
                                                                 nx=nx, ny=ny)
        image.data[:, -noverscan:] = np.random.normal(expected_overscan, expected_read_noise,
                                                      (ny, noverscan))
    images = subtractor.do_stage(images)
    for image in images:
        assert np.abs(image.header['OVERSCAN'][0] - expected_overscan) < 1.0
        assert np.abs(np.mean(image.data[:, :-noverscan]) - input_level + expected_overscan) < 1.0

# TODO: Add test for 2d overscan subtractor

# TODO: Add test for 3d overscan subtractor

# TODO: Copy 2d tests for 3d data cubes



import os
from kombu import Connection, Queue, Exchange
from glob import glob
from astropy.io import fits
import numpy as np
import hashlib

from banzai import dbs
from banzai import logs
from banzai.images import Image


__author__ = 'cmccully'

logger = logs.get_logger(__name__)


def post_to_archive_queue(image_path):
    def errback(exc, interval):
        logger.error('Error: %r', exc, exc_info=1)
        logger.info('Retry in %s seconds.', interval)
    fits_exchange = Exchange('fits_files', type='fanout')
    producer_queue = Queue('pipeline', fits_exchange, exclusive=True)
    with Connection('amqp://guest:guest@cerberus.lco.gtn') as conn:
        queue = conn.SimpleQueue(producer_queue)
        put = conn.ensure(queue, queue.put, max_retries=30, errback=errback)
        put({'path': image_path})


def get_bpm(image, pipeline_context):
    bpm_filename = dbs.get_bpm(image.telescope_id, image.ccdsum,
                               db_address=pipeline_context.db_address)
    if bpm_filename is None:
        bpm_data = np.zeros((image.ny, image.nx), dtype=np.uint8)
        image.header['L1IDMASK'] = ('', 'Id. of mask file used')
    else:
        bpm_data = fits.getdata(bpm_filename)
        image.header['L1IDMASK'] = (os.path.basename(bpm_filename), 'Id. of mask file used')

    return np.array(bpm_data, dtype=np.uint8)


def read_images(image_list, pipeline_context):
    images = []
    for filename in image_list:
        try:
            image = Image(filename=filename)
            if image.bpm is None:
                image.bpm = get_bpm(image, pipeline_context).astype(np.uint8)
            images.append(image)
        except Exception as e:
            logger.error('Error loading {0}'.format(filename))
            logger.error(e)
            continue
    return images


def save_images(pipeline_context, images, master_calibration=False):
    for image in images:
        output_directory = make_output_directory(pipeline_context, image)
        if not master_calibration:
            image.filename = image.filename.replace('00.fits',
                                                    '{:02d}.fits'.format(int(pipeline_context.rlevel)))

        image_filename = os.path.basename(image.filename)
        filepath = os.path.join(output_directory, image_filename)
        image.writeto(filepath, pipeline_context.fpack)
        if pipeline_context.fpack:
            image_filename += '.fz'
            filepath += '.fz'
        if master_calibration:
            dbs.save_calibration_info(image.obstype, filepath, image,
                                      db_address=pipeline_context.db_address)

        if pipeline_context.post_to_archive:
            logger.info('Posting {filename} to the archive'.format(filename=image_filename))
            try:
                post_to_archive_queue(filepath)
            except Exception as e:
                logger.error("Could not post {0} to ingester.".format(filepath))
                logger.error(e)
                continue


def make_image_list(pipeline_context):

    search_path = os.path.join(pipeline_context.raw_path)

    if pipeline_context.filename is None:
        # return the list of file and a dummy image configuration
        fits_files = glob(search_path + '/*.fits')
        fz_files = glob(search_path + '/*.fits.fz')

        fz_files_to_remove = []
        for i, f in enumerate(fz_files):
            if f[:-3] in fits_files:
                fz_files_to_remove.append(i)
        fz_files_to_remove.sort(reverse=True)

        for i in fz_files_to_remove:
            fz_files.pop(i)
        image_list = fits_files + fz_files

    else:
        image_list = glob(os.path.join(pipeline_context.raw_path, pipeline_context.filename))
    return image_list


def select_images(image_list, image_types):
    images = []
    for filename in image_list:
        try:
            if os.path.splitext(filename)[1] == '.fz':
                ext = 1
            else:
                ext = 0
            if fits.getval(filename, 'OBSTYPE', ext=ext) in image_types:
                images.append(filename)
        except Exception as e:
            logger.error('Unable to get OBSTYPE from {0}.'.format(filename))
            logger.error(e)
            continue

    return images


def make_output_directory(pipeline_context, image_config):
    # Create output directory if necessary
    output_directory = os.path.join(pipeline_context.processed_path, image_config.site,
                                    image_config.instrument, image_config.epoch)

    if pipeline_context.preview_mode:
        os.path.join('preview')
    else:
        os.path.join('processed')

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    return output_directory


def get_md5(filepath):
    with open(filepath, 'rb') as file:
        md5 = hashlib.md5(file.read()).hexdigest()
    return md5



from banzai.stages import Stage
from banzai import logs
import os
import numpy as np


class GainNormalizer(Stage):
    def __init__(self, pipeline_context):
        super(GainNormalizer, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for image in images:

            logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            logs.add_tag(logging_tags, 'gain', image.gain)
            self.logger.info('Multiplying by gain', extra=logging_tags)

            if len(image.data.shape) > 2:
                n_amps = image.data.shape[0]
                gain = np.array(eval(image.gain))
                for i in range(n_amps):
                    image.data[i] *= gain[i]
                image.header['SATURATE'] *= min(gain)
                image.header['MAXLIN'] *= min(gain)
            else:
                image.data *= image.gain
                image.header['SATURATE'] *= image.gain
                image.header['MAXLIN'] *= 1.0

            image.gain = 1.0
            image.header['GAIN'] = 1.0
        return images
